// DownConvert444YUV.cpp : Defines the entry point for the console application.
//

#include <string.h>
#include <stdio.h>
#include <stdlib.h>
#include <assert.h>

#define Char	    char      
#define Bool	    bool      
#define Short	    short      
#define UChar	    unsigned char      /* unsigned 8 bit integer	(FF hex)*/
#define UShort    unsigned short int /* unsigned 16 bit integer (FFFF hex)*/
#define UInt	    unsigned int       /* unsigned 32 bit integer	(FFFFFFFF hex)*/
#define Int	      signed   int       /* signed 32 bit integer	(80000000 hex) */

#define NORMALMODE  0   //0 to activate 4 chroma phase filters


#define ROTRS( exp, retVal )  \
{                             \
  if( ( exp ) )               \
  {                           \
    return retVal;            \
  }                           \
}

#define  MAX(x,y)   ((x)>(y)?(x):(y))
#define  MIN(x,y)   ((x)<(y)?(x):(y))

typedef struct
{
  int            stride;
  int            lines;
  unsigned char* data;
} ColorComponent;

typedef struct
{
  ColorComponent y;
  ColorComponent u;
  ColorComponent v;
} YuvFrame;

int
Clip( int iValue, int imin, int imax )
{
  ROTRS( iValue < imin, imin );
  ROTRS( iValue > imax, imax );
  return iValue;
}

void
downsamplingChromaSVC( Int *pTmpFrame, unsigned char*    pucBufferU, unsigned char*    pucBufferV,
                      int   iStride, int frameWidth, int frameHeight, int chromaPhaseX, int chromaPhaseY)
{

//change chroma phase using interpolation filter
  Int iHeight = frameHeight;
  Int iWidth = frameWidth;

  Int minval = 0;
  Int maxval = 255;

  Short lumaFilterQuarter[4][12] = { 
        {   0,  -4,  -7,  11,  38,  52,  38,  11,  -7,  -4,   0,   0 },//phase 0
        {   0,  -2,  -8,   5,  32,  50,  43,  18,  -4,  -8,   2,   0 },//phase 1/4
        {   0,   0,  -8,   0,  24,  48,  48,  24,   0,  -8,   0,   0 }, //phase 1/2
        {   0,   2,  -8,  -4,  18,  43,  50,  32,   5,  -8,  -2,   0 } //phase 3/4
  };
#if NORMALMODE
  Int iHfilt = chromaPhaseX<<1;
  Int iVfilt = chromaPhaseY<<1;
#else
  Int iHfilt = chromaPhaseX;
  Int iVfilt = chromaPhaseY;
#endif

  //memset(pTmpFrame,0,frameHeight * frameWidth/2 * sizeof(Int));

  //FILE *pf =fopen("infodev.txt","at");

  for(Int chromaId=0; chromaId<2 ; chromaId++)
  {
    unsigned char* pChromaAddr = chromaId ? pucBufferV : pucBufferU;

    //horizontal filter
    Int *pTab = pTmpFrame;
    unsigned char* pCAddr = pChromaAddr;

    for(Int i=0; i<iHeight; i++)
    {
      for(Int j=0; j<iWidth>>1; j++) // ss-ech horiz par 2
      {
        Int sum = 0;

        for (Int k=0; k<12; k++)
        {
          Int dj = MAX(0, MIN(iWidth-1, 2*j + k - 5));
          sum +=  pCAddr[dj] * lumaFilterQuarter[iHfilt][k];
        }
        pTab[j] = sum;
   //     fprintf(pf,"%3d ",pTab[j]);
      }
   //   fprintf(pf,"\n");
      pTab += iWidth/2;
      pCAddr += iStride;
    }

   // fprintf(pf,"\n");

    //vertical filter
    pTab = pTmpFrame;
    for(Int j=0; j<iWidth>>1; j++)
    {
      pCAddr = pChromaAddr;
      for(Int i=0; i<iHeight>>1; i++) // ss-ech vertic par 2
      {
        Int sum = 0;

        for (Int k=0; k<12; k++)
        {
          Int di = MAX(0, MIN(iHeight-1, 2*i + k - 5));
          sum +=  pTab[di*iWidth/2+j] * lumaFilterQuarter[iVfilt][k];
        }
        pCAddr[j] = Clip( ((sum + (1<<13) ) >> 14), minval, maxval );
    //    fprintf(pf,"%3d ",pCAddr[j]);
        pCAddr+=iStride;
      }
    //  fprintf(pf,"\n");
    }
    //fprintf(pf,"\n");
    //fclose(pf);
  }
}


void
createColorComponent( ColorComponent& c, int maxwidth, int maxheight )
{
  maxwidth  = ( ( maxwidth  + 15 ) >> 4 ) << 4;
  maxheight = ( ( maxheight + 15 ) >> 4 ) << 4;
  int size  = maxwidth * maxheight;
  c.stride  = maxwidth;
  c.lines   = maxheight;
  c.data    = new unsigned char [ size ];

  if( ! c.data )
  {
    fprintf(stderr, "\nERROR: memory allocation failed!\n\n");
    exit(1);
  }
}

void
deleteColorComponent( ColorComponent& c )
{
  delete[] c.data;
  c.stride  = 0;
  c.lines   = 0;
  c.data    = 0;
}

int
readColorComponent( ColorComponent& c, FILE* file, int width, int height )
{
  assert( width  <= c.stride );
  assert( height <= c.lines  );

  int iMaxPadWidth  = MIN( c.stride, ( ( width  + 15 ) >> 4 ) << 4 );
  int iMaxPadHeight = MIN( c.lines,  ( ( height + 31 ) >> 5 ) << 5 );

  for( int i = 0; i < height; i++ )
  {
    unsigned char* buffer = c.data + i * c.stride;
    int            rsize  = (int)fread( buffer, sizeof(unsigned char), width, file );
    if( rsize != width )
    {
      return 1;
    }
    for( int xp = width; xp < iMaxPadWidth; xp++ )
    {
      buffer[xp] = buffer[xp-1];
    }
  }
  for( int yp = height; yp < iMaxPadHeight; yp++ )
  {
    unsigned char* buffer  = c.data + yp * c.stride;
    unsigned char* bufferX = buffer - c.stride;
    for( int xp = 0; xp < c.stride; xp++ )
    {
      buffer[xp] = bufferX[xp];
    }
  }
  return 0;
}

void
writeColorComponent( ColorComponent& c, FILE* file, int width, int height )
{
  assert( width  <= c.stride );
  assert( height <= c.lines  );

  for( int i = 0; i < height; i++ )
  {
    unsigned char* buffer = c.data + i * c.stride;
    int            wsize  = (int)fwrite( buffer, sizeof(unsigned char), width, file );

    if( wsize != width )
    {
      fprintf(stderr, "\nERROR: while writing to output file!\n\n");
      exit(1);
    }
  }
}

void
createFrame( YuvFrame& f, int width, int height)
{
  createColorComponent( f.y, width , height );
  createColorComponent( f.u, width , height );
  createColorComponent( f.v, width , height );
}

void
deleteFrame( YuvFrame& f )
{
  deleteColorComponent( f.y );
  deleteColorComponent( f.u );
  deleteColorComponent( f.v );
}

int
readFrame( YuvFrame& f, FILE* file, int width, int height )
{
  ROTRS( readColorComponent( f.y, file, width, height ), 1 );
  ROTRS( readColorComponent( f.u, file, width, height ), 1 );
  ROTRS( readColorComponent( f.v, file, width, height ), 1 );
  return 0;
}

void
writeFrame( YuvFrame& f, FILE* file, int width, int height )
{
  writeColorComponent( f.y, file, width,      height );
  writeColorComponent( f.u, file, width >> 1, height >> 1 );
  writeColorComponent( f.v, file, width >> 1, height >> 1 );
}

void
print_usage_and_exit( int test, const char* name, const char* message = 0 )
{
  if( test )
  {
    if( message )
    {
      fprintf ( stderr, "\nERROR: %s\n", message );
    }
    fprintf (   stderr, "\nUsage: %s <win> <hin> <in> <out> [-phase <args>]\n\n");
    fprintf (   stderr, "  win     : input width  (luma samples)\n" );
    fprintf (   stderr, "  hin     : input height (luma samples)\n" );
    fprintf (   stderr, "  in      : input file\n" );
    fprintf (   stderr, "  out     : output file\n" );
    fprintf (   stderr, " -phase <uv_ph_x> <uv_ph_y>\n");
    fprintf (   stderr, "   uv_ph_x: horizontal chroma phase shift, in 1/4th pel (default: 0)\n" );
    fprintf (   stderr, "   uv_ph_y: vertical chroma phase shift, in 1/4th pel   (default: 0)\n" );
    fprintf (   stderr, "\n");
    fprintf (   stderr, "\n\n");
    exit    (   1 );
  }
}


int
main( int argc, char *argv[] )
{
  //===== set standard resize parameters =====
  Int chromaPhaseX = 0;
  Int chromaPhaseY = 0;
  Int frameWidth   = 0;
  Int frameHeight  = 0;

  //===== init parameters =====
  FILE* inputFile  = 0;
  FILE* outputFile = 0;

  //===== read input parameters =====
  printf("\n444 YUV to 420 YUV converter with chroma phase - CANON - S. Pautet - 2011.12.19\n\n");

  print_usage_and_exit( ( argc < 5 || argc > 24 ), argv[0], "wrong number of arguments" );
  frameWidth   = atoi  ( argv[1] );
  frameHeight  = atoi  ( argv[2] );
  inputFile    = fopen ( argv[3], "rb" );
  outputFile   = fopen ( argv[4], "wb" );
  print_usage_and_exit( ! inputFile,  argv[0], "failed to open input file" );
  print_usage_and_exit( ! outputFile, argv[0], "failed to open input file" );

  for( int i = 5; i < argc; )
  {
    if( ! strcmp( argv[i], "-phase" ) )
    {
      i++;
      chromaPhaseX         = atoi( argv[i++] );
      chromaPhaseY         = atoi( argv[i++] );
#if NORMALMODE
      print_usage_and_exit( chromaPhaseX         < 0 || chromaPhaseX         > 1, argv[0], "wrong phase x parameters (range : [0, 1])");
      print_usage_and_exit( chromaPhaseY         < 0 || chromaPhaseY         > 1, argv[0], "wrong phase y parameters (range : [0, 1])");
#endif
    }
    else
    {
      print_usage_and_exit( true, argv[0], "error in command line parameters" );
    }
  }

  //===== initialize classes =====
  YuvFrame    cFrame;
  createFrame( cFrame, frameWidth, frameHeight );

  //===== loop over frames =====
  int   writtenFrames     = 0;
  bool  bFinished         = false;

  Int *pTmpFrame = (Int *) malloc( frameHeight * frameWidth/2 * sizeof(Int));

  while( ! bFinished )
  {
    //===== read input frame =====
    bFinished = (readFrame( cFrame, inputFile, frameWidth, frameHeight) != 0);

    //===== downsampling frame =====
    downsamplingChromaSVC(pTmpFrame, cFrame.u.data, cFrame.v.data,cFrame.u.stride,frameWidth, frameHeight, chromaPhaseX, chromaPhaseY);

    if(!bFinished)
    {
     //===== write output frame =====
      writeFrame( cFrame, outputFile, frameWidth, frameHeight );
      writtenFrames ++;
      fprintf( stderr, "\r%6d frames converted", writtenFrames );
    }
  }

  free(pTmpFrame);
  deleteFrame( cFrame );
  fclose     ( inputFile );
  fclose     ( outputFile );

  return 0;
}

