// dpxToYUV.cpp : Defines the entry point for the console application.
//

#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include "dpxsgiToYUV.h"

UInt clipping(UInt ipix)
{
  int opix;

  if     (ipix>255) opix = 255;
  else if(ipix<0)   opix = 0;
  else              opix = nint(ipix);

  return (opix);
}


void YUV2RGB(UInt Y, UInt U, UInt V, UInt *R, UInt *G, UInt *B)
{
  *R = (UInt)(1.000*Y - 0*U + 1.40*V); 
  *G = (UInt)(1.000*Y - 0.395*U - 0.581*V); 
  *B = (UInt)(1.000*Y + 2.032*U + 0*V);

}
void RGB2YUV(UShort R, UShort G, UShort B, UChar *Y, UChar *U, UChar *V)
{
  *Y = clipping((UInt)( 0.299*R + 0.587*G + 0.114*B) );
  *U = clipping((UInt)(-0.147*R - 0.289*G + 0.436*B + 128));
  *V = clipping((UInt)( 0.615*R - 0.515*G - 0.1*B   + 128));

}

UShort Release(UInt *data, UChar bitdepth)
{
	UInt temp;
  UShort outdata;
	
	outdata = 0;

	temp = (*data )& Detector;
	outdata = temp >>(32-bitdepth); 
	(*data) = (*data) << bitdepth;
	return outdata;
}


int free_frame(YUVimage frame)
{
  free(frame.Y);
  free(frame.U);
  free(frame.V);
  
  return(0);
}

void frame_alloc(YUVimage_ptr frame, Int yWidth, Int yHeight, Int cWidth, Int cHeight)
{
  frame->Y = (UChar *)malloc(yWidth*yHeight * sizeof(UChar));
  frame->U = (UChar *)malloc(cWidth*cHeight * sizeof(UChar));
  frame->V = (UChar *)malloc(cWidth*cHeight * sizeof(UChar));
}


UShort *read_dpx(char *dpxname, struct dpx **dpxheader, Int *yheight, Int *ywidth)
{
  Int nn;
  Int i, j, pos;
  FILE *fp = NULL;
  UInt *data = NULL, rgbU32, offset;
  UShort *RGBframe;
  UChar bitdepth;

  if ((fp = fopen(dpxname, "rb")) == NULL) {
    fprintf(stderr, "Can't open %s for read\n", dpxname);
    exit(1);
  }
  if(fread(*dpxheader, sizeof(struct dpx), 1, fp) != 1){
	  printf("read dpx file header error \n");
	  exit(1);
  }

  *ywidth = (*dpxheader)->imageinfo.pixels_per_line;
  *yheight = (*dpxheader)->imageinfo.lines_per_image_ele;
  offset = (*dpxheader)->fileinfo.offset; 
  //printf("offset = %d\n", offset);

  nn = (*yheight) * (*ywidth);

  fseek(fp, offset, SEEK_SET);

  data = (UInt *)malloc(nn * sizeof(UInt));
  if(fread(data, sizeof(UInt), nn, fp) != (unsigned)nn){
    printf("read image data error\n");
    exit(1);
  }
  fclose(fp);

  /*  seperate 4 byte data into R, G, B */
  bitdepth = (*dpxheader)->imageinfo.bit_size;
  RGBframe = (UShort *)malloc(nn * 3 * sizeof(UShort));

  for(i=0; i<(*yheight); i++){
	  for(j=0; j<(*ywidth); j++){
		  pos = i*(*ywidth) + j;

          rgbU32 = data[pos];

		  RGBframe[i*3*(*ywidth)+3*j+2] = Release(&rgbU32, bitdepth);

		  RGBframe[i*3*(*ywidth)+3*j+1] = Release(&rgbU32, bitdepth);

		  RGBframe[i*3*(*ywidth)+3*j] = Release(&rgbU32, bitdepth);
	  }
  }
  free(data);
  return RGBframe;
}

Short bytestream_memShort(Short &x)
{
  x = ((x<<8) | ((x & 0xff00)>>8) );
  return x;
}
UShort bytestream_memUShort(UShort &x)
{
  x = ((x<<8) | ((x & 0xff00)>>8) );
  return x;
}

UShort *read_sgi(char *sginame, struct sgi **sgiheader, UShort *height, UShort *width, Int *bitdepth)
{
  Int nn;
  Int i, j, pos;
  FILE *fp = NULL;
  UShort *data; 
  UShort *dataR; 
  UShort *dataG; 
  UShort *dataB; 
  UShort rgbU32;
  UShort *RGBframe;

  if ((fp = fopen(sginame, "rb")) == NULL) {
    fprintf(stderr, "Can't open %s for read\n", sginame);
    exit(1);
  }
  if(fread(*sgiheader, sizeof(struct sgi), 1, fp) != 1){
	  printf("read sgi file header error \n");
	  exit(1);
  }

  //info header
  bytestream_memShort((*sgiheader)->Magic);
  if((*sgiheader)->Magic != SGI_MAGIC)
  {
    printf("read sgi file header error : Wrong Magic\n");
    exit(1);
  }

  UInt rle = (*sgiheader)->Storage;
  UInt bytePerChannel = (*sgiheader)->Bpc;
  UInt dimension = bytestream_memShort((*sgiheader)->Dimension);
  *width = bytestream_memUShort((*sgiheader)->XSize);
  *height = bytestream_memUShort((*sgiheader)->YSize);
  Int depth = bytestream_memUShort((*sgiheader)->ZSize);

  if(dimension != 2 && dimension != 3)
  {
    printf("read sgi file error : wrong dimension number\n");
    exit(1);
  }

  if(rle)
  {
    printf("read sgi file error : rle not implemented\n");
    exit(1);
  }

  if(depth!=SGI_RGB)
  {
    printf("read sgi file error : only RGB is implemented\n");
    exit(1);
  }
  *bitdepth = bytePerChannel == 2? 16 : 8;

  nn = (*width) * (*height);

//  fseek(fp, offset, SEEK_SET);

  data = (UShort *)malloc(nn * 3 * sizeof(UShort));
  if(fread(data, sizeof(UShort), 3*nn, fp) != (unsigned)(3*nn)){
    printf("read image data error\n");
    exit(1);
  }
  fclose(fp);

  /*  seperate 4 byte data into R, G, B */
  RGBframe = (UShort *)malloc(nn * 3 * sizeof(UShort));
	pos = 0;

  dataR=data;
  dataG=data+ nn;
  dataB=data+ 2*nn;
  pos = 0;
  Int pos2 = 3*(nn-1);

  for(i=*height-1; i>=0; i--){
	  for(j=*width-1; j>=0; j--){

      rgbU32 = dataR[pos];
      RGBframe[pos2] = bytestream_memUShort(rgbU32);
      rgbU32 = dataG[pos];
      RGBframe[pos2+1] = bytestream_memUShort(rgbU32);
      rgbU32 = dataB[pos++];
      RGBframe[pos2+2] = bytestream_memUShort(rgbU32);
      pos2-=3;
    }
  }

  free(data);
  return RGBframe;
}


void sgi2yuv(char *sginame, char *outname)
{
  UInt pos, pos2;
  UShort ywidth, yheight, cwidth, cheight;
  Int i, j;
  Int bitdepth;
  UShort *RGBframe;
  struct sgi *sgiheader;
  YUVimage YUV;

  sgiheader = (struct sgi *)malloc(sizeof(struct sgi));

  RGBframe = read_sgi(sginame, &sgiheader, &yheight, &ywidth, &bitdepth);

  cwidth = ywidth;
  cheight = yheight;

  frame_alloc(&YUV, ywidth, yheight, cwidth, cheight);

  Int shift = bitdepth-8;

  for(i=0; i<yheight; i++){
	  for(j=0; j<ywidth; j++){
		  pos = i*(ywidth*3)+j*3; /* 3 comes from R, G, B */
		  pos2 = i*ywidth+j;
      RGB2YUV(RGBframe[pos]>>shift, RGBframe[pos+1]>>shift,RGBframe[pos+2]>>shift, &(YUV.Y[pos2]), &(YUV.U[pos2]), &(YUV.V[pos2]));
 	  }
  }

  FILE * pOutFile = fopen(outname,"ab");
  fwrite(YUV.Y, 1, sizeof(UChar)*yheight*ywidth, pOutFile);
  fwrite(YUV.U, 1, sizeof(UChar)*yheight*ywidth, pOutFile);
  fwrite(YUV.V, 1, sizeof(UChar)*yheight*ywidth, pOutFile);
  fclose(pOutFile);

  
  free_frame(YUV); 
  free(sgiheader);
  free(RGBframe);
  return;
}


void dpx2yuv(char *dpxname, char *outname)
{
  Int pos, pos2;
  Int ywidth, yheight, cwidth, cheight;
  Int i, j;
  UShort *RGBframe;
  struct dpx *dpxheader;
  YUVimage YUV;
  //videoinfo info;

  dpxheader = (struct dpx *)malloc(sizeof(struct dpx));

  RGBframe = read_dpx(dpxname, &dpxheader, &yheight, &ywidth);

  ywidth = ywidth;
  yheight = yheight;
  cwidth = ywidth;
  cheight = yheight;

  frame_alloc(&YUV, ywidth, yheight, cwidth, cheight);

  Int bitdepth = dpxheader->imageinfo.bit_size;
  Int shift = bitdepth-8;

  for(i=0; i<yheight; i++){
	  for(j=0; j<ywidth; j++){
		  pos = i*(ywidth*3)+j*3; /* 3 comes from R, G, B */
		  pos2 = i*ywidth+j;
      RGB2YUV(RGBframe[pos+2]>>shift, RGBframe[pos+1]>>shift,RGBframe[pos]>>shift, &(YUV.Y[pos2]), &(YUV.U[pos2]), &(YUV.V[pos2]));
 	  }
  }

  FILE * pOutFile = fopen(outname,"ab");
  fwrite(YUV.Y, 1, sizeof(UChar)*yheight*ywidth, pOutFile);
  fwrite(YUV.U, 1, sizeof(UChar)*yheight*ywidth, pOutFile);
  fwrite(YUV.V, 1, sizeof(UChar)*yheight*ywidth, pOutFile);
  fclose(pOutFile);

  
  free_frame(YUV); 
  free(dpxheader);
  free(RGBframe);
  return;
}

int main(int argc, char* argv[])
{
  int currFrame = 0;
  int firstIdx =0, lastIdx = 0;

  char inputFileName[200];
  char *pInputFileName = &inputFileName[0];
  char currDir[200];
  char *pcurrDir = &currDir[0];
  char outputFileName[200];
  char *pOutputFileName = &outputFileName[0];
  char currFile[300];
  char *pcurrFile = &currFile[0];

  FORMAT format = DPX;  //default value

  printf("\n444 DPX/SGI to 444 YUV converter - CANON - S. Pautet - 2011.12.19\n\n");
  
  printf("Help:\n\n");

  printf("\t-f : input image format \n");
  printf("\t-i : input file prefix \n");
  printf("\t-fidx : first index number \n");
  printf("\t-lidx : last index number \n");
  printf("\t-o : output file name \n");
  printf("\t-d : input files directory \n\n\n");

  if (argc < 2)
    return 0;

  for (int i=1; i < argc; i+=2)
  {
      if(strcmp(argv[i], "-i") == 0)
      {
          strcpy(pInputFileName, argv[i+1]);
          strcpy(pOutputFileName, argv[i+1]);
          strcat(pOutputFileName,"_444.yuv");
      }
      else if(strcmp(argv[i], "-f")==0)
      {
          if(strcmp(argv[i+1], "DPX")==0)
            format = DPX;
          else if(strcmp(argv[i+1], "SGI")==0)
            format = SGI;
      }
      else if(strcmp(argv[i], "-o")==0)
          strcpy(pOutputFileName, argv[i+1]);
      else if(strcmp(argv[i], "-d")==0)
          strcpy(pcurrDir, argv[i+1]);
      else if(strcmp(argv[i], "-fidx")==0)
          firstIdx = atoi(argv[i+1]);
      else if(strcmp(argv[i], "-lidx")==0)
          lastIdx = atoi(argv[i+1]);
  }

  printf("====Input File \t\t: %s \n", pInputFileName);
  printf("====firstIdx  \t\t: %d \n", firstIdx);
  printf("====lastIdx  \t\t: %d \n", lastIdx);
  printf("====Output File \t: %s \n\n", pOutputFileName);

  //Input file
  FILE *pOutFile = fopen(pOutputFileName,"wb");
  fclose (pOutFile);

  
  currFrame = firstIdx;

  if(format==DPX)
  {
    while(currFrame <= lastIdx)
    {
      sprintf(currFile,"%s\\%s%d.dpx",pcurrDir,pInputFileName,currFrame);
      printf("current file : %s%d\n", pInputFileName,currFrame);

      dpx2yuv(currFile, pOutputFileName);
      currFrame++;
    }
  }
  else if(format==SGI)
  {
    while(currFrame <= lastIdx)
    {
      sprintf(currFile,"%s\\%s%d.sgi",pcurrDir,pInputFileName,currFrame);
      printf("current file : %s%d\n", pInputFileName,currFrame);

      sgi2yuv(currFile, pOutputFileName);
      currFrame++;
    }
  }

	return 0;
}
